from itertools import groupby

from teloclip.motifs import check_sequence_for_patterns
from teloclip.utils import isfile


def makeMask(killIdx, listlen):
    # makeMask([0,9], 10) =  [0,1,1,1,1,1,1,1,1,0]
    mask = [1 for i in range(listlen)]
    for x in killIdx:
        mask[x] = 0
    return mask


# NCU
def filterList(data, exclude):
    # filterList([1,2,3,4,5,6,7,8,9,10],[0,9]) = [2,3,4,5,6,7,8,9]
    mask = makeMask(exclude, len(data))
    return (d for d, s in zip(data, mask) if s)


# NCU
def revComp(seq):
    """Rev comp DNA string."""
    revcompl = lambda x: "".join(
        [{"A": "T", "C": "G", "G": "C", "T": "A", "N": "N"}[B] for B in x][::-1]
    )
    return revcompl(seq)


# NCU
def writeClip(idx, zpad, gap, seq, maplen):
    # leftpad idx ID
    padIdx = str(idx).zfill(zpad) + ":"
    # If gap between aln end(R) or start(L) and contig end, left pad softclip with '-'
    padseq = "-" * gap + seq
    # Format length of ref covered by alingment
    readlen = "LEN=" + str(maplen).rjust(6)
    print("\t".join([padIdx, readlen, padseq]))


# NCU
def fasta2dict(fasta_name):
    fh = open(fasta_name)
    faiter = (x[1] for x in groupby(fh, lambda line: line[0] == ">"))
    contigDict = dict()
    for header in faiter:
        # Drop the ">"
        # Split on whitespace and take first item as name
        header = header.__next__()[1:].strip()
        name = header.split()[0]
        # Join all sequence lines to one.
        seq = "".join(s.strip() for s in faiter.__next__())
        contigDict[name] = (header, seq)
    return contigDict


def writefasta(outfile, name, seq, length=80):
    outfile.write(">" + str(name) + "\n")
    while len(seq) > 0:
        outfile.write(seq[:length] + "\n")
        seq = seq[length:]


# NCU
def manageTemp(record=None, tempPath=None, scrub=False):
    """Create single sequence fasta files or scrub temp files."""
    if scrub and tempPath:
        try:
            os.remove(tempPath)
        except OSError:
            pass
    else:
        with open(tempPath, "w") as f:
            name, seq = record
            writefasta(f, name, seq, length=80)


def read_fai(fai):
    """
    Import fasta index file. Return dict of sequence names and lengths.
    """
    path = isfile(fai)
    # Init empty dict
    ContigDict = dict()
    # Read fai_file to dict
    with open(path, "r") as f:
        for line in f.readlines():
            li = line.strip().split()
            ContigDict[li[0]] = int(li[1])
    return ContigDict


def addRevComplement(motifList):
    """
    Take list of DNA motif strings and return unique set of strings and their reverse complements.
    """
    revcompl = lambda x: "".join(
        [{"A": "T", "C": "G", "G": "C", "T": "A", "N": "N"}[B] for B in x][::-1]
    )
    setList = list()
    for motif in motifList:
        setList.append(motif)
        setList.append(revcompl(motif))
    return set(setList)


# Depreciated
def crunchHomopolymers(motifList):
    """
    Take as input a list of target motifs, collapse poly-nucleotide tracks, return list of collapsed motifs.
    """
    # List to catch all collapsed motifs.
    crunchList = list()
    # For each motif
    for motif in motifList:
        # Create empty list to catch not repeated bases.
        noReps = list()
        # Walk through original motif base-by-base.
        for base in motif:
            # If list of kept bases in empty, add first base.
            if not noReps:
                noReps.append(base)
            # If list exists and base is not same as last, store new base.
            elif base != noReps[-1]:
                noReps.append(base)
        # Convert list to string and store new motif
        crunchList.append("".join(noReps))
    # Convert to set to remove duplicates and return
    return list(set(crunchList))


# TODO: support min pattern matches
def isMotifInClip(samline, motifList, leftClip, rightClip, leftClipLen, rightClipLen):
    """
    Extract terminal soft-clipped blocks from read sequence and test for presence of any DNA motif in motifList.
    """
    # Sam seq field
    SAM_SEQ = 9

    # Initialize leftcheck and rightcheck
    leftcheck = False
    rightcheck = False

    # Search motif/s as regex in the clipped segment
    if leftClip:
        leftcheck = check_sequence_for_patterns(
            samline[SAM_SEQ][0:leftClipLen], motifList
        )
    if rightClip:
        rightcheck = check_sequence_for_patterns(
            samline[SAM_SEQ][-rightClipLen:], motifList
        )

    # True if either clipped end sequence contains at least one instance of any motif.
    return any([leftcheck, rightcheck])



import pytest
from src.teloclip.motifs import (
    count_continuous_runs,
    construct_regex_pattern,
    format_pattern_counts,
    count_patterns_in_sequence,
    check_sequence_for_patterns,
)


def test_check_sequence_for_patterns():
    # Test case with a simple DNA sequence and patterns
    dna_sequence = "ATCGATCGATCG"
    regex_patterns = ["ATC", "GAT", "CGA"]
    assert check_sequence_for_patterns(dna_sequence, regex_patterns) is True

    # Test case with an empty DNA sequence
    assert check_sequence_for_patterns("", ["ATC", "GAT"]) is False

    # Test case with an empty pattern list
    assert check_sequence_for_patterns("ATCG", []) is False

    # Test case with no matches in the DNA sequence
    dna_sequence = "ATCGATCGATCG"
    regex_patterns = ["AAA", "GGG", "CCC"]
    assert check_sequence_for_patterns(dna_sequence, regex_patterns) is False

    # Test case with a complex DNA sequence and patterns
    dna_sequence = "ATCGGATCGAGCGCGAATCG"
    regex_patterns = ["ATC", "GAT", "CGA"]
    assert check_sequence_for_patterns(dna_sequence, regex_patterns) is True


def test_count_patterns_in_sequence():
    # Test case with a simple DNA sequence and patterns
    dna_sequence = "ATCGATCGATCG"
    regex_patterns = ["ATC", "GAT", "CGA"]
    assert count_patterns_in_sequence(dna_sequence, regex_patterns) == {
        "ATC": 3,
        "GAT": 2,
        "CGA": 2,
    }

    # Test case with an empty DNA sequence
    assert count_patterns_in_sequence("", ["ATC", "GAT"]) == {"ATC": 0, "GAT": 0}

    # Test case with an empty pattern list
    assert count_patterns_in_sequence("ATCG", []) == {}

    # Test case with a complex DNA sequence and patterns
    dna_sequence = "ATCGGATCGAGCGCGAATCG"
    regex_patterns = ["ATC", "GAT", "CGA"]
    assert count_patterns_in_sequence(dna_sequence, regex_patterns) == {
        "ATC": 3,
        "GAT": 1,
        "CGA": 2,
    }


def test_format_pattern_counts():
    # Test case with a simple pattern counts dictionary
    pattern_counts = {"ATC": 3, "GAT": 1, "CGA": 2}
    assert format_pattern_counts(pattern_counts) == "ATC=3:GAT=1:CGA=2"

    # Test case with an empty pattern counts dictionary
    assert format_pattern_counts({}) == ""

    # Test case with a single pattern count
    assert format_pattern_counts({"ATC": 1}) == "ATC=1"

    # Test case with patterns and counts containing special characters
    pattern_counts = {"A(TC)": 2, "G+AT": 3, "C.A": 1}
    assert format_pattern_counts(pattern_counts) == "A(TC)=2:G+AT=3:C.A=1"


def test_count_continuous_runs():
    # Test case with a simple sequence
    assert count_continuous_runs("AAATTTGGCCCG") == [
        ("A", 3),
        ("T", 3),
        ("G", 2),
        ("C", 3),
        ("G", 1),
    ]

    # Test case with an empty string
    assert count_continuous_runs("") == []

    # Test case with a single character
    assert count_continuous_runs("A") == [("A", 1)]

    # Test case with alternating characters
    assert count_continuous_runs("ATATAT") == [
        ("A", 1),
        ("T", 1),
        ("A", 1),
        ("T", 1),
        ("A", 1),
        ("T", 1),
    ]


def test_construct_regex_pattern():
    # Test case with a simple motif
    motif_sequence = "TTTAGGG"
    motif_tuples = count_continuous_runs(motif_sequence)
    assert construct_regex_pattern(motif_tuples) == r"T{2,4}AG{2,4}"

    # Test case with a motif of length 1
    motif_sequence = "A"
    motif_tuples = count_continuous_runs(motif_sequence)
    assert construct_regex_pattern(motif_tuples) == r"A"

    # Test case with an empty motif
    assert construct_regex_pattern([]) == ""

    # Test case with a motif of alternating characters
    motif_sequence = "ATATAT"
    motif_tuples = count_continuous_runs(motif_sequence)
    assert construct_regex_pattern(motif_tuples) == r"ATATAT"

    # Additional test case with a more complex motif
    motif_sequence = "AAAAGGGTTTCCCC"
    motif_tuples = count_continuous_runs(motif_sequence)
    assert construct_regex_pattern(motif_tuples) == r"A{3,5}G{2,4}T{2,4}C{3,5}"


# Run the tests
if __name__ == "__main__":
    pytest.main()



"""
                                       ██████╗ ██╗      ██╗ ██████╗
                                      ██╔════╝ ██║      ██║ ██╔══██╗
                                      ██║      ██║      ██║ ██████╔╝
████████╗ ███████╗ ██╗       ██████╗  ██║      ██║      ██║ ██╔═══╝
╚══██╔══╝ ██╔════╝ ██║      ██╔═══██╗ ╚██████╗ ███████╗ ██║ ██║
   ██║    █████╗   ██║      ██║   ██║  ╚═════╝ ╚══════╝ ╚═╝ ╚═╝
   ██║    ██╔══╝   ██║      ██║   ██║
   ██║    ███████╗ ███████╗ ╚██████╔╝
   ╚═╝    ╚══════╝ ╚══════╝  ╚═════╝

A tool for the recovery of unassembled telomeres from soft-clipped read alignments.

"""

import argparse
import logging
import sys

from teloclip._version import __version__
from teloclip.logs import init_logging
from teloclip.samops import processSamlines
from teloclip.seqops import read_fai, addRevComplement


def mainArgs():
    parser = argparse.ArgumentParser(
        description="Filter SAM file for clipped alignments containing unassembled telomeric repeats.",
        prog="teloclip",
    )
    # Input options
    parser.add_argument(
        "samfile", nargs="?", type=argparse.FileType("r"), default=sys.stdin
    )
    parser.add_argument(
        "--refIdx",
        type=str,
        required=True,
        help="Path to fai index for reference fasta. Index fasta using `samtools faidx FASTA`",
    )
    parser.add_argument(
        "--minClip",
        type=int,
        default=1,
        help="Require clip to extend past ref contig end by at least N bases.",
    )
    parser.add_argument(
        "--maxBreak",
        type=int,
        default=50,
        help="Tolerate max N unaligned bases before contig end.",
    )
    parser.add_argument(
        "--motifs",
        type=str,
        default=None,
        help="If set keep only reads containing given motif/s from comma delimited list of strings. \
                        By default also search for reverse complement of motifs. i.e. TTAGGG,TTAAGGG will also match CCCTAA,CCCTTAA",
    )
    parser.add_argument(
        "--noRev",
        default=False,
        action="store_true",
        help="If set do NOT search for reverse complement of specified motifs.",
    )
    parser.add_argument(
        "--fuzzy",
        default=False,
        action="store_true",
        help="If set, tolerate +/- 1 variation in motif homopolymer runs \
            i.e. TTAGGG -> T{1,3}AG{2,4}. Default: Off",
    )
    parser.add_argument(
        "-r",
        "--min_repeats",
        default=3,
        type=int,
        help="Minimum number of sequential pattern matches required for a hit to be reported. Default: 3",
    )
    parser.add_argument(
        "--noPoly",
        default=False,
        action="store_true",
        help="WARNING: OPTION DEPRECIATED, USE --FUZZY. Default: False",
    )
    parser.add_argument(
        "--matchAnywhere",
        default=False,
        action="store_true",
        help="If set, motif match may occur in unclipped region of alignment.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = parser.parse_args()
    return args


def main():
    # Set up logging
    init_logging()

    # Get cmd line args
    args = mainArgs()

    # Fetch contig lens
    ContigDict = read_fai(args.refIdx)

    # Split motifs into list
    if args.motifs:
        motifList = args.motifs.split(",")
        # Add rev comp motifs to list, make unique set.
        if not args.noRev:
            motifList = addRevComplement(motifList)
        if args.fuzzy:
            motifList = motifList  # TODO: Implement convert to regex
            # convert to regex pattern(motifList)
    else:
        motifList = []

    # Log depreciation warning if using --noPoly option
    if args.noPoly:
        logging.warning(
            "WARNING: Option --noPoly is depreciated and should not be used. \
            Switching to option --fuzzy instead to find inexact motif matches."
        )
        args.fuzzy = True

    if args.fuzzy:
        logging.info(
            "INFO: Using option --fuzzy to find inexact motif matches. \
                Tolerate +/- 1 base variance in motif homopolymers."
        )

    processSamlines(
        args.samfile,
        ContigDict,
        motifList,
        matchAnywhere=args.matchAnywhere,
        maxBreak=args.maxBreak,
        minClip=args.minClip,
        noRev=args.noRev,
        fuzzy=args.fuzzy,
        minRepeats=args.min_repeats,
    )



import argparse
import logging
import os
import sys

from teloclip._version import __version__
from teloclip.logs import init_logging
from teloclip.samops import StreamingSamFilter
from teloclip.seqops import writefasta, read_fai

"""
Module under development. Currently splits overhang reads of each contig end into separate fata files. 
Users can then assemble these reads externally and realign to the ref contig to repair missing telomeres.
"""


def mainArgs():
    parser = argparse.ArgumentParser(
        description="Extract overhanging reads for each end of each reference contig. Write to fasta.",
        prog="teloclip-extract",
    )

    # Input options
    parser.add_argument(
        "samfile", nargs="?", type=argparse.FileType("r"), default=sys.stdin
    )

    parser.add_argument(
        "--refIdx",
        type=str,
        required=True,
        help="Path to fai index for reference fasta. Index fasta using `samtools faidx FASTA`",
    )

    parser.add_argument(
        "--prefix",
        type=str,
        default=None,
        required=False,
        help="Use this prefix for output files. Default: None.",
    )

    parser.add_argument(
        "--extractReads",
        default=False,
        action="store_true",
        help="If set, write overhang reads to fasta by contig.",
    )

    parser.add_argument(
        "--extractDir",
        type=str,
        default=None,
        required=False,
        help="Write extracted reads to this directory. Default: cwd.",
    )

    parser.add_argument(
        "--minClip",
        type=int,
        default=1,
        help="Require clip to extend past ref contig end by at least N bases.",
    )

    parser.add_argument(
        "--maxBreak",
        type=int,
        default=50,
        help="Tolerate max N unaligned bases before contig end.",
    )

    # Version info
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )

    args = parser.parse_args()

    return args


def StreamingSplitByContig(alignments=None, contigs=None, prefix=None, outdir=None):
    """
    Takes alignment line summaries tagged with overhang information from StreamingSamFilter
    Writes reads into output files for each end of each contig with at least one alignment found.
    """
    # Note: This method opens and closes the output fasta files for EVERY read processed.
    # Probably inefficient but avoids reading everything into memory
    if outdir:
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
    else:
        outdir = os.getcwd()
    # Output file tracker
    outpaths = list()
    # Counter
    readCount = 0
    # For each aligned read
    for read in alignments:
        readCount += 1
        # Log update every 10K reads
        if not readCount % 10000:
            logging.info("Alignments processed: %s" % str(readCount))
        # Check if alignment is at right end overhang
        if read[6] == "R":
            # i.e [(alnStart,alnEnd,rightClipLen,readSeq,readName,contigName,'R')]
            if prefix:
                base = "_".join([prefix, str(read[5])])
            else:
                base = str(read[5])
            # Set output file path
            outfileR = os.path.join(outdir, "_".join([base, "R"]) + ".fasta")
            # Check if need to create new outfile or append to existing file.
            if outfileR not in outpaths:
                logging.info("Creating new outfile: %s" % str(outfileR))
                outpaths.append(outfileR)
                # Output reads aligned to right end of contig
                with open(outfileR, "w") as fileR:
                    # Convert overhanging segment of read to lowercase
                    masked = read[3][: -read[2]] + read[3][-read[2] :].lower()
                    # Write masked read to fasta
                    writefasta(fileR, str(read[4]), masked)
            else:
                # If file already exists
                # Output reads aligned to right end of contig
                with open(outfileR, "a") as fileR:
                    # Convert overhanging segment of read to lowercase
                    masked = read[3][: -read[2]] + read[3][-read[2] :].lower()
                    # Write masked read to fasta
                    writefasta(fileR, str(read[4]), masked)
        # Check if alignment is a left end overhang
        elif read[6] == "L":
            # i.e [(alnStart,alnEnd,leftClipLen,readSeq,readname,contigname,'L')]
            if prefix:
                base = "_".join([prefix, str(read[5])])
            else:
                base = str(read[5])
            # Set output file path
            outfileL = os.path.join(outdir, "_".join([base, "L"]) + ".fasta")
            # Check if need to create new outfile or append to existing file.
            if outfileL not in outpaths:
                logging.info("Creating new outfile: %s" % str(outfileL))
                outpaths.append(outfileL)
                # Output reads aligned to left end of contig
                with open(outfileL, "w") as fileL:
                    # Convert overhanging segment of read to lowercase
                    masked = read[3][: read[2]].lower() + read[3][read[2] :]
                    # Write masked read to fasta
                    writefasta(fileL, str(read[4]), masked)
            else:
                # If file already exists
                # Output reads aligned to left end of contig
                with open(outfileL, "a") as fileL:
                    # Convert overhanging segment of read to lowercase
                    masked = read[3][: read[2]].lower() + read[3][read[2] :]
                    # Write masked read to fasta
                    writefasta(fileL, str(read[4]), masked)

    logging.info("Total alignments processed: %s" % str(readCount))


def main():
    # Set up logging
    init_logging()

    # Get cmd line args
    args = mainArgs()

    # Load ref contigs lengths as dict
    logging.info("Importing reference contig info from: %s " % str(args.refIdx))
    contigInfo = read_fai(args.refIdx)

    # Load alignments from samfile or stdin
    # {'contig':{"L":[(alnStart,alnEnd,leftClipLen,readSeq,readname)],"R":[(alnStart,alnEnd,rightClipLen,readSeq,readname)]}}
    logging.info("Processing alignments. Searching for overhangs.")

    alignments = StreamingSamFilter(
        samfile=args.samfile,
        contigs=contigInfo,
        maxBreak=args.maxBreak,
        minClip=args.minClip,
    )

    if args.extractReads:
        logging.info("Writing overhang reads by contig.")
        # splitbycontig(alignments=alignments,contigs=contigInfo,prefix=args.prefix,outdir=args.extractDir)
        StreamingSplitByContig(
            alignments=alignments,
            contigs=contigInfo,
            prefix=args.prefix,
            outdir=args.extractDir,
        )

    sys.exit(0)



import re
from typing import List, Tuple, Dict


def format_pattern_counts(pattern_counts: Dict[str, int]) -> str:
    """
    Format a dictionary of pattern counts into a string.

    Parameters:
    - pattern_counts (Dict[str, int]): A dictionary where keys are regex patterns
                                    and values are counts of each pattern.

    Returns:
    - str: A formatted string with pattern counts.
    """
    # Use a list comprehension to generate formatted pairs of pattern=count
    formatted_pairs = [
        f"{pattern}={count}" for pattern, count in pattern_counts.items()
    ]

    # Join the formatted pairs with ":" and return the result
    return ":".join(formatted_pairs)


def count_patterns_in_sequence(
    dna_sequence: str, regex_patterns: List[str]
) -> Dict[str, int]:
    """
    Count occurrences of regular expression patterns in a DNA sequence.

    Parameters:
    - dna_sequence (str): The input DNA sequence.
    - regex_patterns (List[str]): List of regular expression patterns to count.

    Returns:
    - Dict[str, int]: A dictionary where keys are regex patterns and values are
                     counts of each pattern in the sequence.
    """
    pattern_counts = {}  # Initialize an empty dictionary to store counts

    # Iterate through the list of regex patterns
    for pattern in regex_patterns:
        # Use re.findall to find all occurrences of the pattern in the DNA sequence
        matches = re.findall(pattern, dna_sequence)
        # Count the number of matches and store in the dictionary
        pattern_counts[pattern] = len(matches)

    # Or as a dict comprehension
    # pattern_counts = {pattern: len(re.findall(pattern, dna_sequence)) for pattern in regex_patterns}

    return pattern_counts


def check_sequence_for_patterns(dna_sequence: str, regex_patterns: List[str]) -> bool:
    """
    Check a DNA sequence for instances of one or more regular expressions.

    Parameters:
    - dna_sequence (str): The input DNA sequence.
    - regex_patterns (List[str]): List of regular expression patterns to check against.

    Returns:
    - bool: True if any of the patterns match the sequence, False otherwise.
    """
    # Use any() to check if any pattern matches the DNA sequence
    return any(re.search(pattern, dna_sequence) for pattern in regex_patterns)


def count_continuous_runs(dna_string: str) -> list:
    """
    Count the length of all continuous runs of a DNA base in a string.

    Parameters:
    - dna_string (str): The input DNA string.

    Returns:
    - list: A list of tuples where each tuple contains a character and the number
            of times it occurred consecutively in the input DNA string.
    """

    # Check if the input string is empty
    if not dna_string:
        return []

    # Initialize variables
    current_base = dna_string[0]
    current_count = 1
    result = []

    # Iterate through the DNA string starting from the second character
    for base in dna_string[1:]:
        if base == current_base:
            # If the current base is the same as the previous one, increment the count
            current_count += 1
        else:
            # If the current base is different, append the tuple and reset the count
            result.append((current_base, current_count))
            current_base = base
            current_count = 1

    # Append the last tuple after the loop
    result.append((current_base, current_count))

    return result


def construct_regex_pattern(motif_tuples: List[Tuple[str, int]]) -> str:
    """
    Construct a regex pattern to match sequences with runs of characters
    that differ by plus or minus one compared to the original input sequence.

    Parameters:
    - motif_tuples (List[Tuple[str, int]]): List of tuples where each tuple contains
                                            a character and the number of times it
                                            occurred consecutively.

    Returns:
    - str: The constructed regex pattern as a raw string.
    """

    pattern_parts = []

    for char, count in motif_tuples:
        if count == 1:
            pattern_parts.append(
                re.escape(char)
            )  # If count is 1, just escape the character
        else:
            # If count is greater than 1, add a range allowing for plus or minus one
            pattern_parts.append(rf"{re.escape(char)}{{{count-1},{count+1}}}")

    return rf"{''.join(pattern_parts)}"



import logging
import sys


def init_logging():
    # Define the log message format
    fmt = "%(asctime)s | %(levelname)s | %(module)s:%(lineno)s:%(funcName)s() | %(message)s"

    # Set up a StreamHandler for stderr instead of stdout
    handler_sh = logging.StreamHandler(sys.stderr)
    handler_sh.setFormatter(CustomFormatter(fmt))

    # Configure the logging with our handler for stderr
    logging.basicConfig(format=fmt, level=logging.DEBUG, handlers=[handler_sh])


class CustomFormatter(logging.Formatter):
    """Logging colored formatter, adapted from https://alexandra-zaharia.github.io/posts/make-your-own-custom-color-formatter-with-python-logging"""

    grey = "\x1b[38;21m"
    blue = "\x1b[38;5;39m"
    yellow = "\x1b[38;5;226m"
    red = "\x1b[38;5;196m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    def __init__(self, fmt):
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.DEBUG: self.grey + self.fmt + self.reset,
            logging.INFO: self.blue + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset,
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)
